from __future__ import annotations

import os
import pickle
from dataclasses import dataclass, field
from datetime import datetime
from pathlib import Path
from typing import Optional, Any, Union

import numpy as np
import yaml

from slimfit import Model
from slimfit.utils import clean_types


@dataclass
class FitResult:
    """

    """

    parameters: dict[str, float | np.ndarray]
    """Fitted parameter values"""

    gof_qualifiers: dict
    """Goodness-of-fit qualifiers"""

    fixed_parameters: Optional[dict[str, float]] = None
    """Values of the model's fixed parameters"""

    guess: Optional[dict] = None
    """Initial guesses"""

    symbolic_model: Optional[Model] = None
    """The fitted symbolic model"""

    data: Optional[dict] = field(default=None, repr=False)
    """Data on which the fit was performed"""

    metadata: dict = field(default_factory=dict)
    """Additional metadata"""

    base_result: Optional[Any] = field(default=None, repr=False)
    """Source fit result object. Can be dicts of sub results"""

    def __post_init__(self) -> None:
        if "datetime" not in self.metadata:
            now = datetime.now()
            self.metadata["datetime"] = now.strftime("%Y/%m/%d %H:%M:%S")
            self.metadata["timestamp"] = int(now.timestamp())

    def to_dict(self) -> dict:
        keys = ["gof_qualifiers", "parameters", "fixed_parameters", "guess", "metadata"]
        d = {k: v for k in keys if (v := getattr(self, k)) is not None}

        return d

    def to_yaml(self, path: Union[os.PathLike[str], str], sort_keys: bool = False) -> None:
        """
        Save the fitresult to yaml
        """

        dic = clean_types(self.to_dict())
        Path(path).write_text(yaml.dump(dic, sort_keys=sort_keys))

    def to_pickle(self, path: Union[os.PathLike[str], str]) -> None:
        """
        Save the fitresult as pickle

        Args:
            path: Path to save to.
        """

        with Path(path).open("wb") as f:
            pickle.dump(self, f)

    def __call__(self, **kwargs) -> dict[str, np.ndarray]:
        raise NotImplementedError("Nope")
        data = self.data or {}
        kwargs = self.parameters | data | kwargs

        return self.symbolic_model(**kwargs)



from __future__ import annotations

from typing import TypeVar, KeysView, ValuesView, ItemsView, Protocol

from typing_extensions import TypeVarTuple

# PEP 646: https://peps.python.org/pep-0646/
Shape = TypeVarTuple("Shape")

DataType = TypeVar("DataType")


# STRATEGY / PROTOCOL?
# class Data(UserDict, Generic[DataType]):
class Data(Protocol[DataType]):

    ...

    def __getitem__(self, item) -> DataType:
        return super().__getitem__(item)

    def __setitem__(self, key: str, value: DataType) -> None:
        return super().__setitem__(key, value)

    def keys(self) -> KeysView[str]:
        return super().keys()

    def values(self) -> ValuesView[DataType]:
        return super().values()

    def items(self) -> ItemsView[str, DataType]:
        return super().items()


if __name__ == "__main__":
    import numpy as np

    d = {"a": np.random.rand(10, 20)}
    data: Data[np.ndarray] = d

    print(data)



from __future__ import annotations

import typing

import numpy as np
import sympy as sp
from sympy import Matrix, zeros, Symbol, Expr

from slimfit import MatrixNumExpr

if typing.TYPE_CHECKING:
    from slimfit.symbols import Variable


def gaussian_numpy(
    x: np.array | float, mu: np.array | float, sig: np.array | float
) -> np.array | float:
    return 1 / (np.sqrt(2 * np.pi) * sig) * np.exp(-np.power((x - mu) / sig, 2) / 2)


def gaussian_sympy(x: Symbol, mu: Symbol, sig: Symbol) -> Expr:
    return sp.exp(-((x - mu) ** 2) / (2 * sig ** 2)) / (sp.sqrt(2 * np.pi) * sig)


def gaussian(
    x: Variable, mu: Matrix | MatrixNumExpr, sigma: Matrix | MatrixNumExpr
) -> Matrix | MatrixNumExpr:
    if mu.shape != sigma.shape:
        raise ValueError("Shape mismatch between 'mu' and 'sigma'")

    m_out = zeros(*mu.shape)
    for i, j in np.ndindex(mu.shape):
        m_out[i, j] = gaussian_sympy(x, mu[i, j], sigma[i, j])

    if isinstance(mu, MatrixNumExpr) or isinstance(sigma, MatrixNumExpr):
        return MatrixNumExpr(m_out, kind="GMM")
    else:
        return m_out


def gaussian_matrix(x: Variable, mu: Matrix, sigma: Matrix) -> Matrix:
    """Applies gaussian function elementwise from a 'mu' and 'sigma' matrix; given a variable 'x'"""
    if mu.shape != sigma.shape:
        raise ValueError("Shape mismatch between 'mu' and 'sigma'")

    m_out = zeros(*mu.shape)
    for i, j in np.ndindex(mu.shape):
        m_out[i, j] = gaussian_sympy(x, mu[i, j], sigma[i, j])

    return m_out



from __future__ import annotations

import abc
from functools import cached_property

from sympy import Symbol
import numpy as np

from slimfit.parameter import Parameter, Parameters
from slimfit.typing import Shape


class SymbolicBase(metaclass=abc.ABCMeta):
    @cached_property
    @abc.abstractmethod
    def symbols(self) -> set[Symbol]:
        ...

    @cached_property
    def symbol_names(self) -> set[str]:
        return set(s.name for s in self.symbols)

    @property
    def shapes(self) -> dict[str, Shape]:
        """
        dict of symbol shapes
        """

    def filter_parameters(self, parameters: Parameters) -> Parameters:
        """Filters a list of parameters, returning only the ones whose symbols are
        in this model
        """
        return Parameters([p for p in parameters if p.symbol in self.symbols])

    # @property
    # @abc.abstractmethod
    # def shape(self) -> tuple:
    #     ...



from __future__ import annotations

import itertools
from functools import cached_property, reduce
from operator import or_
from typing import (
    Union,
    Callable,
    Optional,
    Any,
    Mapping,
    Iterable,
    TYPE_CHECKING,
    KeysView,
    ItemsView,
    ValuesView,
)

import numpy as np
from scipy.integrate import solve_ivp
from sympy import Expr, MatrixBase, lambdify, HadamardProduct, Matrix, Symbol

from slimfit.base import SymbolicBase

# from slimfit.base import SymbolicBase
from slimfit.parameter import Parameters, Parameter
from slimfit.symbols import FitSymbol
from slimfit.typing import Shape

if TYPE_CHECKING:
    from slimfit import Model

# @dataclass
class NumExprBase(SymbolicBase):
    """Symbolic expression which allows calling cached lambified expressions
    subclasses must implement `symbols` attribute / property
    """

    # def __init__(
    #     self,
    # ):
    #
    #     # Accepted parameters are a subset of `symbols`
    #     # #todo property with getter / setter where setter filters parameters?
    #     # self.parameters = Parameters({name: p for name, p in parameters.items() if name in self.symbols})

    @property
    def shape(self) -> Shape:
        shapes = self.shapes.values()

        return np.broadcast_shapes(*shapes)

    def parse_kwargs(self, **kwargs) -> dict[str, np.ndarray]:
        """Parse kwargs and take only the ones in `free_parameters`"""
        try:
            arguments: dict[str, np.ndarray | float] = {k: kwargs[k] for k in self.symbol_names}
        except KeyError as e:
            print(kwargs.keys())
            raise KeyError(f"Missing value for parameter {e}") from e

        return arguments


class DummyNumExpr(NumExprBase):
    """Dummy callable object which returns supplied 'obj' when called
    Has no parameters or symbols
    """

    def __init__(self, obj: Any, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.obj = obj

    def __call__(self, **kwargs):
        return self.obj

    @property
    def symbols(self) -> set[Symbol]:
        return set()


class ArrayNumExpr(DummyNumExpr):
    @property
    def shape(self) -> Shape:
        return self.obj.shape


# TODO frozen dataclass?
class NumExpr(NumExprBase):
    def __init__(
        self, expr: Expr,
    ):
        if not isinstance(expr, (Expr, MatrixBase)):
            # TODO subclass such that typing is correct
            raise TypeError(f"Expression must be an instance of `Expr` or ")
        self.expr = expr

        # super().__init__()

    @property
    def name(self) -> str:
        # if _name: ... -> superclass
        if isinstance(self.expr, Symbol):
            return self.expr.name

    # is same as callablematrix
    @property
    def symbols(self) -> set[Symbol]:
        return self.expr.free_symbols

    @cached_property
    def lambdified(self) -> Callable:
        ld = lambdify(sorted(self.symbols, key=str), self.expr)

        return ld

    def __call__(self, **kwargs: float) -> np.ndarray | float:

        # try:
        #     parameters: dict[str, np.ndarray | float] = {
        #         k: kwargs[k] for k in self.parameters.keys()
        #     }
        # except KeyError as e:
        #     raise KeyError(f"Missing value for parameter {e}") from e

        val = self.lambdified(**self.parse_kwargs(**kwargs))
        return val

    def __repr__(self):
        return f"NumExpr({self.expr})"


# todo name via kwargs to super
# = composite num expr"?
class MatrixNumExpr(NumExpr):
    def __init__(
        self, expr: MatrixBase, name: Optional[str] = None, kind: Optional[str] = None,
    ):

        if not isinstance(expr, MatrixBase):
            raise TypeError("Expression must be an instance of MatrixParameter or sympy.Matrix")
        self._name = name

        # Callable type is used by minimizers to determine solving strategy
        if kind is None:
            self.kind = identify_expression_kind(expr)
        elif isinstance(kind, str):
            self.kind: str = kind.lower()
        else:
            raise TypeError("Invalid type for 'kind', must be 'str'")

        super().__init__(expr)

    @property
    def name(self) -> str:
        if self._name:
            return self._name
        if self.kind == "constant":
            symbol_names = [str(symbol) for symbol in self.expr]
            prefix = [name.split("_")[0] for name in symbol_names]
            if len(set(prefix)) == 1:  # All prefixes are identical, prefix is the name
                return prefix[0]
        else:
            return "M"

    @name.setter
    def name(self, name):
        self._name = name

    @property
    def shape(self) -> Shape:
        # again there might be problems depending on how the matrix elements depend on
        # different combinations of parameters and data
        # for now we assume this is the same for all elements

        # Find the shape from broadcasting parameters and data
        base_shape = super().shape

        # squeeze last dim if shape is (1,)
        base_shape = () if base_shape == (1,) else base_shape
        shape = base_shape + self.expr.shape

        return shape

    @cached_property
    def element_mapping(self) -> dict[Symbol, tuple[int, int]]:
        """
        Dictionary mapping Symbol to matrix indices
        only returns entries if the matrix element is equal to a `Symbol`, not `Expr`
        """

        element_mapping = {}
        for i, j in np.ndindex(self.expr.shape):
            elem = self.expr[i, j]
            if isinstance(elem, Symbol):
                element_mapping[elem] = (i, j)
        return element_mapping

    @cached_property
    def lambdified(self) -> np.ndarray:
        """Array of lambdified function per matrix element"""
        # TODO scalercallable per element

        lambdas = np.empty(self.expr.shape, dtype=object)
        # todo might go wrong when not all elements have the same parameters
        for i, j in np.ndindex(self.expr.shape):
            lambdas[i, j] = lambdify(sorted(self.symbols, key=str), self.expr[i, j])

        return lambdas

    def forward(self, **kwargs):
        """forward pass with type checking

        or without?
        """
        ...

    def __call__(self, **kwargs: float) -> np.ndarray:
        # https://github.com/sympy/sympy/issues/5642
        # Prepare kwargs for lambdified
        # try:
        #     parameters: dict[str, np.ndarray | float] = {
        #         k: kwargs[k] for k in self.free_parameters.keys()
        #     }
        # except KeyError as e:
        #     raise KeyError(f"Missing value for parameter {e}") from e

        # check shapes
        # this should move somewhere else
        # for p_name, p_value in parameters.items():
        #     if getattr(p_value, "shape", tuple()) != self.parameters[p_name].shape:
        #         raise ValueError(f"Shape mismatch for parameter {p_name}")

        ld_kwargs = self.parse_kwargs(**kwargs)

        # todo precomputed shapes
        # try:
        #     # Shape is given be pre-specified shape
        #     shape = self.shape
        # except AttributeError:
        base_shape = np.broadcast_shapes(
            *(getattr(value, "shape", tuple()) for value in ld_kwargs.values())
        )

        # squeeze last dim if shape is (1,)
        base_shape = () if base_shape == (1,) else base_shape
        shape = base_shape + self.expr.shape

        out = np.empty(shape)
        for i, j in np.ndindex(self.expr.shape):
            out[..., i, j] = self.lambdified[i, j](**ld_kwargs)

        return out

    @property
    def values(self) -> np.ndarray:
        """

        Returns: Array with elements set to parameter values

        """
        raise DeprecationWarning("Deprecate in favour of `value`")
        arr = np.empty(self.shape)
        for i, j in np.ndindex(self.shape):
            arr[i, j] = self.expr[i, j].value

        return arr

    def __getitem__(self, key):
        return self.expr[key]

    def __contains__(self, item) -> bool:
        return self.expr.__contains__(item)

    # this is more of a str than a repr
    def __repr__(self):
        names = sorted(self.symbol_names)
        return f"{self.name}({', '.join(names)})"

    def index(self, symbol: Symbol) -> tuple[int, int]:
        """
        Returns indices of parameter for Matrix Expressions

        Args:
            name: Parameter name to find matrix elements of

        Returns: Tuple of matrix elements ij

        """

        return self.element_mapping[symbol]


class Constant(MatrixNumExpr):
    # WIP of a class which has an additional variable which determines output shape but
    # is no symbol in underlying expression

    def __init__(self, x: Symbol, m: Matrix, name: Optional[str] = None):
        raise NotImplementedError("Nope")
        self.x = x
        super().__init__(m, name=name)

    def __call__(self, **kwargs):
        m_vals = super().__call__(**kwargs)

        np.broadcast_to(...)


class DummyVariableMatrix(MatrixNumExpr):
    """
    Matrix callable which takes an additional variable such that called returned shape is expanded to accomodate its
    shapes
    """

    def __init__(
        self, x: Symbol, m: Matrix, kind: Optional[str] = None, name: Optional[str] = None,
    ):
        raise NotImplementedError("Not implemented")
        self.x = x
        super().__init__(m, kind=kind, name=name)

    @property
    def symbols(self) -> dict[str, Symbol]:
        symbols = super().symbols | {self.x}
        return symbols


# different class for Symbolic / Numerical ?
class LambdaNumExpr(NumExprBase):
    def __init__(self, func, symbols: Iterable[Symbol],) -> None:
        self.func = func
        self._symbols = set(symbols)

    @property
    def symbols(self):
        return self._symbols

    def __call__(self, **kwargs):
        return self.func(**self.parse_kwargs(**kwargs))


# refactor to Hybrid ?
class CompositeExpr(SymbolicBase):
    """Can be both numerical or symbolic """

    def __init__(
        self, expr: dict[str | Symbol, NumExprBase | Expr | CompositeExpr],
    ):
        self.expr = expr

    def __call__(self, **kwargs) -> dict[str, np.ndarray]:
        return {expr_name: expr(**kwargs) for expr_name, expr in self.expr.items()}

    def __getitem__(self, item) -> NumExprBase | Expr:
        return self.expr.__getitem__(item)

    @property
    def numerical(self) -> bool:
        return all(isinstance(v, (NumExprBase, CompositeExpr)) for v in self.values())

    def keys(self) -> KeysView[str]:
        return self.expr.keys()

    def values(self) -> ValuesView[NumExprBase, Expr]:
        return self.expr.values()

    def items(self) -> ItemsView[str, NumExprBase, Expr]:
        return self.expr.items()

    def to_numerical(self):
        num_expr = {str(k): to_numerical(expr) for k, expr in self.items()}

        instance = self.__class__(num_expr)
        return instance

    @cached_property
    def symbols(self) -> set[Symbol]:
        """Return symbols in the CompositeNumExpr.
        sorting is by dependent_variables, variables, parameters, then by alphabet
        """

        # this fails because `free_symbols` is a dict on NumExpr but `set` on Expr

        symbols = set()
        for rhs in self.values():
            if isinstance(rhs, (Expr, MatrixBase)):
                symbols |= rhs.free_symbols
            else:
                try:
                    symbols |= set(rhs.symbols)
                except AttributeError:
                    # RHS doesnt have any symbols; for example might be a numpy array
                    pass

            # symbols = getattr(rhs, 'free_symbols')
            # try:
            #     # rhs is a sympy `Expr` and has `free_symbols` as a set
            #     symbols |= rhs.free_symbols
            # except TypeError:
            #     # rhs is a slimfit `NumExpr`
            #     symbols |= set(rhs.symbols)
            # except AttributeError:
            #     # RHS doesnt have any symbols; for example might be a numpy array
            #     pass
        return symbols

    @property
    def shapes(self) -> dict[str, Shape]:
        """shapes of symbols"""
        return reduce(or_(expr.shapes for expr in self.expr.values()))

    @property
    def shape(self) -> Shape:
        """
        Base class shape is obtained from broadcasting all expressing values together
        """
        shapes = (expr.shape for expr in self.values())
        return np.broadcast_shapes(*shapes)


class GMM(CompositeExpr):
    # todo can also be implemented as normal NumExpr but with broadcasting parameter shapes
    # important is that GMM class allows users to find oud positions of parmaeters in mu / sigma
    # matrices for EM GMM optimization.

    def __init__(
        self,
        x: Symbol | NumExpr,
        mu: Matrix | MatrixNumExpr,
        sigma: Matrix | MatrixNumExpr,
        name: Optional[str] = None,
    ):
        # check for the correct shape when creating from sympy Matrix
        if isinstance(mu, Matrix) and mu.shape[0] != 1:
            raise ValueError(
                "GMM parameter matrices must be of shape (1, N) where N is the number of states."
            )

        expr = {"x": x, "mu": mu, "sigma": sigma}

        # todo some kind of properties object for this metadata
        name = name or "GMM"  # counter for number of instances?
        self.kind = "gmm"
        super().__init__(expr)

    def __call__(self, **kwargs) -> np.ndarray:
        result = super().__call__(**kwargs)

        x, mu, sig = result["x"], result["mu"], result["sigma"]
        output = 1 / (np.sqrt(2 * np.pi) * sig) * np.exp(-np.power((x - mu) / sig, 2) / 2)

        # Output shape is (datapoints, states, 1) to match output shape of matrix
        # exponentiation model
        return np.expand_dims(output, -1)

    def __repr__(self):
        return f"GMM({self.expr['x']}, {self.expr['mu']}, {self.expr['sigma']})"

    def to_numerical(self) -> GMM:
        # todo probably this is the same as the super class method
        num_expr = {k: to_numerical(expr) for k, expr in self.items()}
        instance = GMM(**num_expr)

        return instance

    @property
    def shape(self) -> Shape:
        shape = super().shape
        return shape + (1,)

    # TODO GMM should have API to find the index of the state given a symbol
    #def state_index(self, ...):

class MarkovIVP(CompositeExpr):
    """Uses scipy.integrate.solve_ivp to numerically find time evolution of a markov process
        given a transition rate matrix.

    Returned shape is (len(t_var), len(y0), 1), or (<datapoints>, <states>, 1)

    """

    def __init__(
        self,
        t: Symbol | NumExpr,
        trs_matrix: Matrix | MatrixNumExpr,
        y0: Matrix | MatrixNumExpr,
        domain: Optional[tuple[float, float]] = None,
        **ivp_kwargs,
    ):

        expr = {"t": t, "trs_matrix": trs_matrix, "y0": y0}

        super().__init__(expr)

        ivp_defaults = {"method": "Radau"}
        self.ivp_defaults = ivp_defaults | ivp_kwargs
        self.domain = domain

    def __call__(self, **kwargs):
        result = super().__call__(**kwargs)

        # if `self['t']` does not depend on any parameters; domain can be precomputed and
        # does not have to be determined for every call
        # although its every fast to do so
        domain = self.domain or self.get_domain(result["t"])
        sol = solve_ivp(
            self.grad_func,
            domain,
            y0=result["y0"].squeeze(),
            t_eval=result["t"],
            args=(result["trs_matrix"],),
            **self.ivp_defaults,
        )

        # shape is modified to match the output shape of matrix exponentiation model
        # exp(m*t) @ y0, which is (datapoints, states, 1)
        return np.expand_dims(sol.y.T, -1)

    def to_numerical(self) -> MarkovIVP:
        num_expr = {k: to_numerical(expr) for k, expr in self.items()}
        instance = MarkovIVP(**num_expr, domain=self.domain, **self.ivp_defaults)

        return instance

    def get_domain(self, arr: np.ndarray) -> tuple[float, float]:
        # padding?
        return arr[0], arr[-1]

    @staticmethod
    def grad_func(t, y, trs_matrix):
        return trs_matrix @ y


def identify_expression_kind(sympy_expression: Union[Expr, MatrixBase]) -> str:
    """Find the type of expression

    Only implemented for 'constant' kind, otherwise return is generic

    """

    # check for gaussian mixture model ...

    if isinstance(sympy_expression, MatrixBase):
        ...

        if all(isinstance(elem, Symbol) for elem in sympy_expression):
            # this should also check the symbols for their shapes, for it to be constant
            # the elements should all be scalars
            return "constant"

    return "generic"


def to_numerical(
    expression: Union[NumExprBase, Expr, MatrixBase | Model | CompositeExpr],
) -> NumExprBase | CompositeExpr:
    """Converts sympy expression to slimfit numerical expression

        if the expressions already is an NumExpr; the object is modified in-place by setting
        the parameters

    """
    from slimfit.models import Model

    if hasattr(expression, "to_numerical"):
        return expression.to_numerical()
    # elif isinstance(expression, Model):
    #     model_dict = {lhs: to_numerical(rhs) for lhs, rhs in expression.items()}
    #     from slimfit.models import NumericalModel
    #
    #     return NumericalModel(model_dict, parameters, data)
    if isinstance(expression, HadamardProduct):
        from slimfit.operations import Mul

        return Mul(*(to_numerical(arg) for arg in expression.args))
    elif isinstance(expression, MatrixBase):
        return MatrixNumExpr(expression)
    elif isinstance(expression, Expr):
        return NumExpr(expression)
    elif isinstance(expression, np.ndarray):
        return ArrayNumExpr(expression)
    elif isinstance(expression, NumExprBase):
        return expression
    else:
        raise TypeError(f"Invalid type {type(expression)!r}")



from __future__ import annotations

from typing import Callable, Union

import numpy as np

ReductionStrategy = Callable[[dict[str, np.array]], Union[float, np.array]]


def mean_reduction(residuals: dict[str, np.array]) -> float:
    size = sum(arr.size for arr in residuals.values())
    total = sum(r.sum() for r in residuals.values())
    return total / size


def sum_reduction(residuals: dict[str, np.array]) -> float:
    return sum(r.sum() for r in residuals.values())


def concat_reduction(residuals: dict[str, np.array]) -> np.array:
    return np.concatenate([a.ravel() for a in residuals.values()])



from functools import reduce
from operator import add
from typing import Optional, Callable, Type

import numpy as np
from sympy import Matrix, zeros, Symbol

OPERATORS = ["<->", "<-", "->"]


def generate_transition_matrix(
    connectivity: list[str],
    parameter_prefix="k",
    check_mass_balance=True,
    symbol_class: Type[Symbol] = Symbol,
):
    all_states = extract_states(connectivity)

    b = ["_" in s for s in all_states]
    if any(b):
        raise ValueError("Underscores are not allowed in state names")

    trs_matrix = zeros(len(all_states), len(all_states))
    for conn in connectivity:
        split = conn.split(" ")
        states = [s for s in split if not s in OPERATORS]

        for current_state in states:
            i = split.index(current_state)
            current_idx = all_states.index(current_state)

            # look to the left
            if i >= 2:
                op = split[i - 1]
                other_state = split[i - 2]  # refactor other
                other_idx = all_states.index(other_state)
                if op in ["->", "<->"]:  # flux from other state to current state
                    # elem = self.create_element(other_state, current_state)
                    elem = symbol_class(f"{parameter_prefix}_{other_state}_{current_state}")
                    trs_matrix[current_idx, other_idx] += elem

                if op in ["<-", "<->"]:  # flux from current state to other state
                    # elem = self.create_element(current_state, other_state)
                    elem = symbol_class(f"{parameter_prefix}_{current_state}_{other_state}")
                    trs_matrix[current_idx, current_idx] -= elem

            # look to the right
            if i <= len(split) - 2:
                op = split[i + 1]
                other_state = split[i + 2]
                other_idx = all_states.index(other_state)

                if op in ["<-", "<->"]:  # flux from other state to current state
                    # elem = self.create_element(other_state, current_state)
                    elem = symbol_class(f"{parameter_prefix}_{other_state}_{current_state}")
                    trs_matrix[current_idx, other_idx] += elem
                if op in ["->", "<->"]:  # flux from current state to other state
                    # elem = self.create_element(current_state, other_state)
                    elem = symbol_class(f"{parameter_prefix}_{current_state}_{other_state}")
                    trs_matrix[current_idx, current_idx] -= elem

    return trs_matrix


def exp_elements(
    m: Matrix, base=10, sub_name: Optional[Callable] = lambda x: "u" + x[1:]
) -> Matrix:
    """
    Exponentiate elements of matrix to a given base, and optionally substitute parameters with new parameters

    Args:
        m: Input matrix which elements to exponentiate.
        base: Base of the exponent.
        sub_name: Function which takes parameter name and returns substitute parameter name.

    Returns:
        Exponentiated matrix

    """
    out = zeros(*m.shape)
    for i, j in np.ndindex(m.shape):
        elem = m[i, j]
        if elem == 0.0:
            continue
        else:
            out[i, j] = base ** elem

    if sub_name is not None:
        subs = [(p, Parameter(name=sub_name(p.name))) for p in t.free_symbols]
        out = out.subs(subs)

    return out


def extract_states(connectivity: list[str]) -> list[str]:
    """
    Args:
        connectivity: List of reaction equations.

    Returns:
        List of states found in all reaction equations.

    """

    # extract states from connectivity list
    all_states = [
        s for s in reduce(add, [eqn.split(" ") for eqn in connectivity]) if s not in OPERATORS
    ]

    # Remove duplicates, keep order
    all_states = list(dict.fromkeys(all_states))

    return all_states



from __future__ import annotations

import abc
import time
from dataclasses import asdict
from functools import reduce, cached_property
from operator import or_
from typing import Optional, Any

import numpy as np
from scipy.optimize import minimize
from sympy import Symbol
from tqdm.auto import trange

from slimfit import Model, NumExprBase
from slimfit.fitresult import FitResult
from slimfit.loss import Loss
from slimfit.objective import ScipyObjective, pack, unpack, ScipyEMObjective

# from slimfit.models import NumericalModel
from slimfit.operations import Mul
from slimfit.parameter import Parameters, Parameter
from slimfit.utils import get_bounds, overlapping_model_parameters

# TODO parameter which needs to be inferred / set somehow
STATE_AXIS = -2

# dataclass?
class Minimizer(metaclass=abc.ABCMeta):
    def __init__(
        self,
        numerical_model: Model,
        parameters: Parameters,
        loss: Loss,
        xdata: dict[str, np.array],
        ydata: dict[str, np.array],
    ):
        if not numerical_model.numerical:
            raise ValueError("The given model should be numerical")

        self.model = numerical_model
        self.loss = loss
        self.xdata = xdata
        self.ydata = ydata
        self.parameters = parameters

    # subset of parameters which are fixed
    @cached_property
    def fixed_parameters(self) -> Parameters:
        return Parameters([p for p in self.parameters if p.fixed])

    @cached_property
    def free_parameters(self) -> Parameters:
        return Parameters([p for p in self.parameters if not p.fixed])

    @property
    def guess(self) -> dict[str, np.ndarray]:
        # todo also on self.parameters.guess
        raise NotImplementedError("Use `free_parameters.guess")

        print("guess on minimizer")
        for p in self.parameters:
            print(p.name, p.fixed)

        return {p.name: np.asarray(p.guess) for p in self.parameters if not p.fixed}

    def get_bounds(self) -> list[tuple[float | None, float | None]] | None:
        bounds = [(p.lower_bound, p.upper_bound) for p in self.free_parameters]

        if all((None, None) == b for b in bounds):
            return None
        else:
            return bounds

    @abc.abstractmethod
    def execute(self, **minimizer_options) -> FitResult:
        ...


class ScipyMinimizer(Minimizer):
    def execute(self, **minimizer_options):
        param_shapes = {p.name: p.shape for p in self.free_parameters}

        objective = ScipyObjective(
            model=self.model,
            loss=self.loss,
            xdata=self.xdata | self.fixed_parameters.guess,
            ydata=self.ydata,
            shapes=param_shapes,
        )

        x = pack(self.free_parameters.guess.values())

        result = minimize(
            objective, x, bounds=self.get_bounds(), options=self.rename_options(minimizer_options)
        )

        gof_qualifiers = {
            "loss": result["fun"],
        }

        parameter_values = unpack(result.x, param_shapes)
        result_dict = dict(
            parameters=parameter_values,
            fixed_parameters=self.fixed_parameters,
            gof_qualifiers=gof_qualifiers,
            guess=self.free_parameters.guess,
            base_result=result,
        )

        # todo pass to superclass generalize fitresult function
        return FitResult(**result_dict)

    def rename_options(self, options: dict[str, Any]) -> dict[str, Any]:
        # todo parse options more generally
        rename = [("max_iter", "maxiter")]

        out = options.copy()
        out.pop("stop_loss", None)
        return out


# should take an optional CompositeNumExpr which returns the posterior
# todo refactor to EM Optimizer?
class LikelihoodOptimizer(Minimizer):
    """
    Assumed `loss` is `LogLoss`

    """

    def execute(self, max_iter=250, patience=5, stop_loss=1e-7, verbose=True) -> FitResult:
        # parameters which needs to be passed / inferred
        # Split top-level multiplications in the model as they can be optimized in log likelihood independently
        components: list[tuple[Symbol, NumExprBase]] = []  # todo tuple LHS as variable
        for lhs, rhs in self.model.items():
            if isinstance(rhs, Mul):
                components += [(lhs, elem) for elem in rhs.values()]
            else:
                components.append((lhs, rhs))

        # Find the sets of components which have common parameters and thus need to be optimized together
        sub_models = overlapping_model_parameters(components, self.free_parameters.symbols)

        pbar = trange(max_iter, disable=not verbose)
        t0 = time.time()

        parameters_current = self.free_parameters.guess  # initialize parameters
        prev_loss = 0.0
        no_progress = 0
        base_result = {}
        for i in pbar:
            y_model = self.model(**parameters_current, **self.xdata, **self.fixed_parameters.guess)
            loss = self.loss(self.ydata, y_model)

            # posterior dict has values with shapes equal to y_model
            # which is (dataopints, states, 1)
            posterior = {k: v / v.sum(axis=STATE_AXIS, keepdims=True) for k, v in y_model.items()}

            # dictionary of new parameter values for in this iteration
            parameters_step = {}
            common_kwargs = dict(
                loss=self.loss, xdata=self.xdata, ydata=self.ydata, posterior=posterior,
            )
            for sub_model in sub_models:
                # At the moment we assume all callables in the sub models to be MatrixCallables
                # determine the kind
                kinds = [c.kind for c in sub_model.values()]
                # Filter the general list of parameters to reduce it down to the parameters
                # this model accepts
                # the sub model also needs to the fixed parameters to correctly be able to
                # call the model; GMM also needs it for finding sigma
                sub_parameters = sub_model.filter_parameters(self.parameters)
                if all(k == "constant" for k in kinds):
                    # Constant optimizer doesnt use loss, xdata, ydata,
                    opt = ConstantOptimizer(sub_model, sub_parameters, **common_kwargs)
                    parameters = opt.step()
                elif all(k == "gmm" for k in kinds):
                    # Loss is not used for GMM optimizer step
                    opt = GMMOptimizer(sub_model, sub_parameters, **common_kwargs)
                    parameters = opt.step()
                else:

                    # Previous code:
                    # updated_parameters = [
                    #     Parameter(**(asdict(p) | {"guess": parameters_current.get(p.name) or self.fixed_parameters.guess[p.name]  }))
                    #     for p in sub_parameters
                    # ]

                    # New; needs improvement as it accesses `_names`
                    current_sub = {
                        k: v for k, v in parameters_current.items() if k in sub_parameters._names
                    }
                    updated_parameters = sub_parameters.update_guess(current_sub)

                    # todo loss is not used; should be EM loss while the main loop uses Log likelihood loss
                    opt = ScipyEMOptimizer(sub_model, updated_parameters, **common_kwargs)

                    scipy_result = opt.execute()
                    parameters = scipy_result.parameters
                    base_result["scipy"] = scipy_result

                # collect parameters of this sub_model into parmaeters dict
                parameters_step |= parameters

            # update for next iteration
            parameters_current = parameters_step

            # loss
            improvement = prev_loss - loss
            prev_loss = loss
            pbar.set_postfix({"improvement": improvement})
            if np.isnan(improvement):
                break
            elif improvement < stop_loss:
                no_progress += 1
            else:
                no_progress = 0

            if no_progress > patience:
                break

        tdelta = time.time() - t0
        gof_qualifiers = {
            "loss": loss,
            "log_likelihood": -loss,
            "n_iter": i + 1,
            "elapsed": tdelta,
            "iter/s": tdelta / (i + 1),
        }

        result = FitResult(
            parameters=parameters_current,
            fixed_parameters=self.fixed_parameters.guess,
            gof_qualifiers=gof_qualifiers,
            guess=self.free_parameters.guess,
            base_result=base_result,
        )

        return result


class EMOptimizer(Minimizer):
    def __init__(
        self,
        numerical_model: Model,
        parameters: list[Parameter],  # Parameters?
        loss: Loss,
        xdata: dict[str, np.array],
        ydata: dict[str, np.array],
        posterior: dict[str, np.array],
    ):
        self.posterior = posterior
        super().__init__(
            numerical_model=numerical_model,
            parameters=parameters,
            loss=loss,
            xdata=xdata,
            ydata=ydata,
        )

    @abc.abstractmethod
    def step(self) -> dict[str, float]:
        return {}

    def execute(self, max_iter=250, patience=5, stop_loss=1e-7, verbose=True) -> FitResult:
        raise NotImplementedError("Not yet")
        pbar = trange(max_iter, disable=not verbose)
        t0 = time.time()

        parameters_current = self.guess  # initialize parameters
        prev_loss = 0.0
        no_progress = 0
        # cache dict?
        for i in pbar:
            eval = self.model(**parameters_current)
            loss = self.loss(self.ydata, eval)

            parameters_step = self.step()

            # update parameters for next iteration
            parameters_current = parameters_step

            # Check improvement of loss
            improvement = prev_loss - loss
            prev_loss = loss
            pbar.set_postfix({"improvement": improvement})
            if np.isnan(improvement):
                break
            elif improvement < stop_loss:
                no_progress += 1
            else:
                no_progress = 0

            if no_progress > patience:
                break

        tdelta = time.time() - t0
        gof_qualifiers = {
            "loss": loss,
            "log_likelihood": -loss,
            "n_iter": i,
            "elapsed": tdelta,
            "iter/s": tdelta / i,
        }

        result = FitResult(
            parameters=parameters_current,
            fixed_parameters=self.model.fixed_parameters,
            gof_qualifiers=gof_qualifiers,
            guess=self.guess,
            # model=self.model,
            # data={**self.xdata, **self.ydata},
        )

        return result


class GMMOptimizer(EMOptimizer):
    """optimizes parameter values of GMM (sub) model
    doesnt use guess directly but instead finds next values for parmater from posterior probabilities

    rhs of model should all be GMM CompositeNumExprs

    """

    # TODO create `step` method which only does one step', execute should be full loop

    def step(self) -> dict[str, float]:
        parameters = {}  # output parameters dictionary

        # All symbols in all 'mu' expressions, then take intersection
        mu_symbols = set.union(*(rhs["mu"].symbols for rhs in self.model.values()))
        mu_symbols &= self.free_parameters.symbols

        # take only the mu symbos in the set of symbols designated as parameters
        mu_parameters = mu_symbols
        # mu_parameters = reduce(
        #     or_, [rhs["mu"].free_parameters.keys() for rhs in self.model.values()]
        # )
        for mu_symbol in mu_parameters:
            num, denom = 0.0, 0.0
            for lhs, gmm_rhs in self.model.items():
                # check if the current mu parameter in this GMM
                if mu_symbol in gmm_rhs["mu"].symbols:
                    col, state_index = gmm_rhs["mu"].index(mu_symbol)
                    T_i = np.take(self.posterior[str(lhs)], state_index, axis=STATE_AXIS)

                    # independent data should be given in the same shape as T_i
                    # which is typically (N, 1), to be sure shapes match we reshape independent data
                    # data = self.xdata[gmm_rhs['x'].name]
                    num += np.sum(T_i * self.xdata[gmm_rhs["x"].name].reshape(T_i.shape))
                    denom += np.sum(T_i)

            parameters[mu_symbol.name] = num / denom

        sigma_symbols = set.union(*(rhs["sigma"].symbols for rhs in self.model.values()))
        sigma_symbols &= self.free_parameters.symbols

        # sigma_parameters = reduce(
        #     or_, [rhs["sigma"].free_parameters.keys() for rhs in self.model.values()]
        # )
        for sigma_symbol in sigma_symbols:
            num, denom = 0.0, 0.0
            # LHS in numerical models are `str` (at the moment)
            for lhs, gmm_rhs in self.model.items():
                # check if the current sigma parameter in this GMM
                if sigma_symbol in gmm_rhs["sigma"].symbols:
                    col, state_index = gmm_rhs["sigma"].index(sigma_symbol)

                    T_i = np.take(self.posterior[str(lhs)], state_index, axis=STATE_AXIS)

                    # Indexing of the MatrixExpr returns elements of its expr
                    mu_name: str = gmm_rhs["mu"][col, state_index].name

                    # Take the corresponding value from the current parameters dict, if its not
                    # there, it must be in the fixed parameters of the model
                    try:
                        mu_value = parameters[mu_name]
                    except KeyError:
                        mu_value = self.fixed_parameters.guess[mu_name]

                    num += np.sum(
                        T_i * (self.xdata[gmm_rhs["x"].name].reshape(T_i.shape) - mu_value) ** 2
                    )

                    denom += np.sum(T_i)

            parameters[sigma_symbol.name] = np.sqrt(num / denom)

        return parameters


class ConstantOptimizer(EMOptimizer):

    """
    model is of form {Probability(px): Matrix[[<parameters>]] ...} where all matrix elements are just scalar parameters.
    """

    def step(self) -> dict[str, float]:
        parameters = {}
        # for p_name in self.model.free_parameters:
        for parameter in self.free_parameters:
            num, denom = 0.0, 0.0
            for lhs, rhs in self.model.items():
                # rhs is of type MatrixNumExpr
                if parameter.symbol in rhs.symbols:
                    # Shapes of RHS matrices is (N_states, 1), find index with .index(name)
                    state_index, _ = rhs.index(parameter.symbol)
                    T_i = np.take(self.posterior[str(lhs)], state_index, axis=STATE_AXIS)
                    num_i, denom_i = T_i.sum(), T_i.size

                    num += num_i
                    denom += denom_i
            parameters[parameter.name] = num / denom

        return parameters


class ScipyEMOptimizer(EMOptimizer):

    # TODO this is an abstract method
    def step(self):
        ...

    def execute(self, **minimizer_options):
        param_shapes = {p.name: p.shape for p in self.free_parameters}

        # x = self.model.parameters.pack(self.guess)
        # x = np.array([self.guess[p_name] for p_name in self.parameter_names])
        # options = {"method": "SLSQP"} | minimizer_options

        objective = ScipyEMObjective(
            model=self.model,
            loss=self.loss,  # not used currently
            xdata=self.xdata | self.fixed_parameters.guess,
            posterior=self.posterior,
            shapes=param_shapes,
        )

        x = pack(self.free_parameters.guess.values())
        options = minimizer_options
        # bounds = self.model.free_parameters.get_bounds()
        # todo what if users wants different bounds to pass to the minimizer?
        # perhaps that should also be passed, same as guess?
        # what about the pack / unpack?
        result = minimize(
            objective,
            x,
            # args=(self.model, self.loss, self.posterior),
            # args=(self.parameter_names, self.xdata, self.posterior, self.model, self.loss,),
            bounds=self.get_bounds(),
            **options
        )

        gof_qualifiers = {
            "loss": result["fun"],
        }

        parameter_values = unpack(result.x, param_shapes)

        result_dict = dict(
            parameters=parameter_values,
            fixed_parameters=self.fixed_parameters,
            gof_qualifiers=gof_qualifiers,
            guess=self.free_parameters.guess,
            base_result=result,
        )

        # todo pass to superclass generalize fitresult function
        # or functional
        return FitResult(**result_dict)

    # TODO duplicate code
    def to_fitresult(self, result) -> FitResult:

        parameters = self.model.parameters.unpack(result.x)

        gof_qualifiers = {
            "loss": result["fun"],
        }

        fit_result = FitResult(
            parameters=parameters,
            gof_qualifiers=gof_qualifiers,
            guess=self.guess,
            base_result=result,
        )

        return fit_result


MIN_PROB = 1e-9  # Minimal probability value (> 0.) to enter into np.log


def minfunc_expectation(
    x: np.ndarray,  # array of parameters
    model: Model,  # parameter names
    loss: Loss,  # currently not used
    posterior: dict,  # posterior probabilities
):
    params = model.parameters.unpack(x)
    probability = model(**params)

    # Todo do this in a `loss`
    expectation = {
        lhs: posterior[lhs] * np.log(np.clip(prob, a_min=MIN_PROB, a_max=1.0))
        for lhs, prob in probability.items()
    }
    # TODO: LOSS / WEIGHTS

    return -sum(r.sum() for r in expectation.values())


def minfunc(
    x: np.ndarray,  # array of parameters
    model: Model,
    loss: Loss,
    dependent_data: dict,  # corresponding measurements; target data
) -> float:

    parameter_values = model.parameters.unpack(x)
    predicted = model(**parameter_values)

    return loss(dependent_data, predicted)



from pathlib import Path

import pytest

# from dont_fret.em_fit.datagen import generate_dataset
from sympy import HadamardProduct, Matrix, exp, Symbol

# import numpy as np

from slimfit.fit import Fit
from slimfit.functions import gaussian, gaussian_sympy, gaussian_numpy
from slimfit.loss import LogLoss

# from slimfit.markov import generate_transition_matrix, extract_states
# from slimfit.minimizer import LikelihoodOptimizer
from slimfit.operations import Mul, MatMul
from slimfit.models import Model
from slimfit.numerical import MatrixNumExpr, NumExpr, GMM, to_numerical, LambdaNumExpr, MarkovIVP
from slimfit.symbols import (
    symbol_matrix,
    clear_symbols,
    get_symbols,
)
from slimfit.loss import LogSumLoss
from slimfit.minimizer import LikelihoodOptimizer
from slimfit.markov import generate_transition_matrix, extract_states
from slimfit.parameter import Parameters, Parameter
import numpy as np

root_dir = Path(__file__).parent.parent


class TestEMBase(object):
    def test_symbol_matrix(self):
        clear_symbols()
        m = symbol_matrix("A", shape=(3, 3))

        assert m.shape == (3, 3)
        elem = m[0, 0]
        assert elem.name == "A_0_0"

        m = symbol_matrix("A", shape=(1, 3), suffix=["a", "b", "c"],)

        elem = m[0, 0]
        assert elem.name == "A_a"

        elem = m[0, 1]
        assert elem.name == "A_b"

        parameters = {
            "A_a": Parameter(m[0, 0], guess=2.0),
            "A_b": Parameter(m[0, 1]),
            "A_c": Parameter(m[0, 2]),
        }

        m_num = to_numerical(m)
        values = {"A_a": 1, "A_b": 2, "A_c": 3.5}

        result = m_num(**values)
        assert result.shape == (1, 3)
        assert np.allclose(result, np.array([1.0, 2.0, 3.5]).reshape(1, 3))

    # @pytest.mark.skip("Old test")
    def test_model(self):
        clear_symbols()

        # Create GMM from sympy operations and Parameters
        suffix = ["1", "2", "3"]
        mu = symbol_matrix(name="mu", shape=(3, 1), suffix=suffix)
        sigma = symbol_matrix(name="sigma", shape=(3, 1), suffix=suffix)

        # elementwise gaussian
        g = gaussian(Symbol("x"), mu, sigma)

        c = symbol_matrix(name="c", shape=(3, 1), suffix=suffix)
        model_dict = {Symbol("p"): HadamardProduct(c, g)}

        model = Model(model_dict)
        num_model = model.to_numerical()
        rhs = next(iter(num_model.values()))
        assert isinstance(rhs, Mul)

        symbols = get_symbols(mu, sigma, c)
        parameters = Parameters.from_symbols(symbols.values())
        print(parameters.guess)

        # Test calling the model
        kwargs = {"x": np.linspace(0, 1, num=100), **parameters.guess}
        res = num_model(**kwargs)

        assert res["p"].shape == (100, 3, 1)

        val = sum(np.trapz(f, kwargs["x"]) for f in res["p"].squeeze().T)
        assert 1 == pytest.approx(val, 1)


class TestNumExpr(object):
    def test_num_expr(self):
        clear_symbols()
        np.random.seed(43)

        x = np.arange(100).reshape(-1, 1)
        data = {"x": x}
        parameters = {
            "a": Parameter(Symbol("a"), guess=np.array([1, 2, 3]).reshape(1, -1)),
            "b": Parameter(Symbol("b"), guess=5.0),
        }

        expr = Symbol("a") * Symbol("x") + Symbol("b")
        num_expr = NumExpr(expr)

        # todo test shapes
        # assert num_expr.shape == (100, 3)

        a = np.random.rand(1, 3)
        b = 5.0
        result = num_expr(a=a, b=b, **data)
        assert np.allclose(result, a * x + b)

    def test_matrix_num_expr(self):
        clear_symbols()

        m = Matrix(
            [
                [
                    Symbol("a") * Symbol("x") + Symbol("b1"),
                    Symbol("a") * Symbol("x") + Symbol("b2"),
                    Symbol("a") * Symbol("x") + Symbol("b3"),
                ]
            ]
        )

        data = {"x": np.arange(100).reshape(-1, 1)}
        symbols = get_symbols(m)

        parameters = {
            "a": Parameter(symbols["a"], guess=np.random.rand(1, 3)),
            "b1": Parameter(symbols["b1"], guess=1.0),
            "b2": Parameter(symbols["b1"], guess=2.0),
            "b3": Parameter(symbols["b1"], guess=3.0),
        }

        m_expr = MatrixNumExpr(m)
        # todo shapes
        # assert m_expr.shape

        p_values = {
            "a": np.array([3, 2, 1]).reshape(1, -1),
            "b1": 2.0,
            "b2": 3.0,
            "b3": 4.0,
        }

        result = m_expr(**p_values, **data)

        # v assert result.shape == m_expr.shape

        check = data["x"] * p_values["a"] + p_values["b1"]
        assert np.allclose(check, result[..., 0, 0])

        check = data["x"] * p_values["a"] + p_values["b2"]
        assert np.allclose(check, result[..., 0, 1])

        check = data["x"] * p_values["a"] + p_values["b3"]
        assert np.allclose(check, result[..., 0, 2])

        # test symbol matrix factory function
        shape = (1, 3)
        states = ["A", "B", "C"]
        c = symbol_matrix(name="c", shape=shape, suffix=states)
        assert c[0, 0] == Symbol("c_A")

        num_c = to_numerical(c)

        assert num_c.kind == "constant"
        assert num_c.name == "c"

    def test_lambda_numexpr(self):
        clear_symbols()
        np.random.seed(43)

        def func(x, a):
            return x ** 2 + a

        data = {"x": np.arange(100)}

        ld = LambdaNumExpr(func, [Symbol("a"), Symbol("x")],)

        # todo shape testing
        # assert ld.shape == (100,)

        result = ld(a=2.0, **data)
        assert np.allclose(result, data["x"] ** 2 + 2.0)

    def test_gmm(self):
        states = ["A", "B", "C"]
        mu = symbol_matrix("mu", suffix=states, shape=(1, 3))
        sigma = symbol_matrix("sigma", suffix=states, shape=(1, 3))
        gmm = GMM(Symbol("x"), mu, sigma)
        parameters = Parameters.from_symbols(gmm.symbols, "mu_A mu_B mu_C sigma_A sigma_B sigma_C")
        data = {"x": np.linspace(-0.2, 1.2, num=25).reshape(-1, 1)}

        gt = {
            "mu_A": 0.23,
            "mu_B": 0.55,
            "mu_C": 0.92,
            "sigma_A": 0.1,
            "sigma_B": 0.1,
            "sigma_C": 0.1,
            "c_A": 0.22,
            "c_B": 0.53,
            "c_C": 0.25,
        }

        num_gmm = gmm.to_numerical()
        assert gmm.kind == "gmm"
        # todo shapes
        # assert num_gmm.shape == (25, 3, 1)
        assert isinstance(num_gmm["mu"], MatrixNumExpr)

        result = num_gmm(**gt, **data)
        assert result.shape == (25, 3, 1)


class TestEMFit(object):
    def test_linear_lstsq(self):
        clear_symbols()
        np.random.seed(43)

        gt = {"a": 0.5, "b": 2.5}

        xdata = np.linspace(0, 11, num=100)
        ydata = gt["a"] * xdata + gt["b"]

        noise = np.random.normal(0, scale=ydata / 10.0 + 0.5)
        ydata += noise

        data = {"x": xdata, "y": ydata}

        model = Model({Symbol("y"): Symbol("a") * Symbol("x") + Symbol("b")})
        parameters = Parameters.from_symbols(model.symbols, "a b")
        fit = Fit(model, parameters, data)
        res = fit.execute()

        assert res.parameters["a"] == pytest.approx(gt["a"], abs=0.2)
        assert res.parameters["b"] == pytest.approx(gt["b"], abs=0.1)

    def test_likelihood_gaussian(self):
        clear_symbols()
        np.random.seed(43)

        gt = {"mu": 2.4, "sigma": 0.7}

        data = {"x": np.random.normal(gt["mu"], scale=gt["sigma"], size=100)}
        model = Model({Symbol("p"): gaussian_sympy(Symbol("x"), Symbol("mu"), Symbol("sigma"))})

        parameters = Parameters.from_symbols(model.symbols, "mu sigma")

        fit = Fit(model, parameters, data, loss=LogLoss())
        res = fit.execute()

        assert res.parameters["mu"] == pytest.approx(gt["mu"], abs=0.05)
        assert res.parameters["sigma"] == pytest.approx(gt["sigma"], abs=0.05)

    def test_linear_matrix(self):
        clear_symbols()
        np.random.seed(43)

        mu_vals = [1.1, 3.5, 7.2]
        sigma_vals = [0.25, 0.1, 0.72]
        wavenumber = np.linspace(0, 11, num=100)  # wavenumber x-axis
        basis = np.stack(
            [gaussian_numpy(wavenumber, mu_i, sig_i) for mu_i, sig_i in zip(mu_vals, sigma_vals)]
        ).T

        x_vals = np.array([0.3, 0.5, 0.2]).reshape(3, 1)  # unknowns
        spectrum = basis @ np.array(x_vals).reshape(3, 1)  # measured

        x = symbol_matrix(name="X", shape=(3, 1))
        symbols = get_symbols(x)
        parameters = Parameters.from_symbols(symbols.values())

        model = Model({Symbol("b"): MatMul(basis, x)})
        fit = Fit(model, parameters, data={"b": spectrum})
        result = fit.execute()

        for i, j in np.ndindex(x_vals.shape):
            assert x_vals[i, j] == pytest.approx(result.parameters[f"X_{i}_{j}"], rel=1e-3)

    def test_exponential_matrix(self):
        clear_symbols()
        np.random.seed(43)

        gt_values = {
            "k_A_B": 1e0,
            "k_B_A": 5e-2,
            "k_B_C": 5e-1,
            "y0_A": 1.0,
            "y0_B": 0.0,
            "y0_C": 0.0,
        }

        connectivity = ["A <-> B -> C"]
        m = generate_transition_matrix(connectivity)
        states = extract_states(connectivity)

        xt = exp(m * Symbol("t"))
        y0 = symbol_matrix(name="y0", shape=(3, 1), suffix=states)
        model = Model({Symbol("y"): xt @ y0})

        rate_params = Parameters.from_symbols(get_symbols(m).values())
        y0_params = Parameters.from_symbols(get_symbols(y0).values())
        parameters = rate_params + y0_params

        num = 50
        xdata = {"t": np.linspace(0, 11, num=num)}
        num_model = model.to_numerical()
        populations = num_model(**gt_values, **xdata)["y"]
        ydata = {
            "y": populations + np.random.normal(0, 0.05, size=num * 3).reshape(populations.shape)
        }

        fit = Fit(model, parameters, data={**xdata, **ydata})
        result = fit.execute()

        expected = {
            "k_A_B": 1.0926495267297978,
            "k_B_A": 0.02553115392319696,
            "k_B_C": 0.48848195581215753,
            "y0_A": 1.0144580699136068,
            "y0_B": -0.011557732388925912,
            "y0_C": -0.006383620511149652,
        }

        for k in expected:
            assert result.parameters[k] == pytest.approx(expected[k], rel=0.1)

    def test_markov_ivp(self):
        clear_symbols()
        np.random.seed(43)

        gt_values = {
            "k_A_B": 1e0,
            "k_B_A": 5e-2,
            "k_B_C": 5e-1,
            "y0_A": 1.0,
            "y0_B": 0.0,
            "y0_C": 0.0,
        }

        connectivity = ["A <-> B -> C"]
        m = generate_transition_matrix(connectivity)
        states = extract_states(connectivity)

        y0 = symbol_matrix(name="y0", shape=(3, 1), suffix=states)
        model = Model({Symbol("y"): MarkovIVP(Symbol("t"), m, y0)})

        rate_params = Parameters.from_symbols(get_symbols(m).values())
        y0_params = Parameters.from_symbols(get_symbols(y0).values())
        parameters = rate_params + y0_params

        num = 50
        xdata = {"t": np.linspace(0, 11, num=num)}

        num_model = model.to_numerical()
        populations = num_model(**gt_values, **xdata)["y"]

        ydata = {
            "y": populations + np.random.normal(0, 0.05, size=num * 3).reshape(populations.shape)
        }
        ydata["y"].shape  # shape of the data is (50, 3, 1)

        fit = Fit(model, parameters, data={**xdata, **ydata})
        result = fit.execute()

        expected = {
            "k_A_B": 1.0928473669526968,
            "k_B_A": 0.025697062159894243,
            "k_B_C": 0.4884426216496761,
            "y0_A": 1.0144463997001603,
            "y0_B": -0.011475371097876597,
            "y0_C": -0.006402646431591074,
        }

        for k in expected:
            assert result.parameters[k] == pytest.approx(expected[k], rel=0.1)

    def test_gmm(self):
        clear_symbols()
        np.random.seed(43)

        gt = {
            "mu_A": 0.23,
            "mu_B": 0.55,
            "mu_C": 0.92,
            "sigma_A": 0.1,
            "sigma_B": 0.1,
            "sigma_C": 0.1,
            "c_A": 0.22,
            "c_B": 0.53,
            "c_C": 0.25,
        }

        np.random.seed(43)
        N = 1000
        states = ["A", "B", "C"]
        xdata = np.concatenate(
            [
                np.random.normal(
                    loc=gt[f"mu_{s}"], scale=gt[f"sigma_{s}"], size=int(N * gt[f"c_{s}"])
                )
                for s in states
            ]
        )

        np.random.shuffle(xdata)
        data = {"x": xdata.reshape(-1, 1)}

        guess = {
            "mu_A": 0.2,
            "mu_B": 0.4,
            "mu_C": 0.7,
            "sigma_A": 0.1,
            "sigma_B": 0.1,
            "sigma_C": 0.1,
            "c_A": 0.33,
            "c_B": 0.33,
            "c_C": 0.33,
        }

        g_shape = (1, 3)
        c_shape = (3, 1)
        mu = symbol_matrix(name="mu", shape=g_shape, suffix=states)
        sigma = symbol_matrix(name="sigma", shape=g_shape, suffix=states)
        c = symbol_matrix(name="c", shape=c_shape, suffix=states)

        model = Model({Symbol("p"): Mul(c, GMM(Symbol("x"), mu, sigma))})

        symbols = get_symbols(mu, sigma, c)
        parameters = Parameters.from_symbols(symbols.values(), guess)

        fit = Fit(model, parameters, data, loss=LogSumLoss(sum_axis=1))
        result = fit.execute(minimizer=LikelihoodOptimizer, verbose=False)

        expected = {
            "c_A": 0.21807944048774222,
            "c_B": 0.5351105112590985,
            "c_C": 0.24681004825315928,
            "mu_A": 0.23155221598099554,
            "mu_B": 0.5508567564172897,
            "mu_C": 0.9204744537231175,
            "sigma_A": 0.09704934271877938,
            "sigma_B": 0.09910459765563108,
            "sigma_C": 0.09877267156818363,
        }

        for k in expected:
            assert result.parameters[k] == pytest.approx(expected[k], rel=0.1)

        # Repeat with fixed parameters

        # fix mu A
        parameters.set("mu_A", fixed=True)
        fit = Fit(model, parameters, data, loss=LogSumLoss(sum_axis=1))
        result = fit.execute(minimizer=LikelihoodOptimizer, verbose=False)

        expected = {
            "c_A": 0.18205289103947245,
            "c_B": 0.5822954009717819,
            "c_C": 0.23565170798874566,
            "mu_B": 0.5427685760300586,
            "mu_C": 0.9273517146087585,
            "sigma_A": 0.08436777980742473,
            "sigma_B": 0.11294214298924928,
            "sigma_C": 0.09474026756245091,
        }
        for k in expected:
            assert result.parameters[k] == pytest.approx(expected[k], rel=0.1)

        for fixed_param in ["mu_A"]:
            assert result.fixed_parameters[fixed_param] == guess[fixed_param]

        # fix sigma B
        parameters.set("sigma_B", fixed=True)
        fit = Fit(model, parameters, data, loss=LogSumLoss(sum_axis=1))
        result = fit.execute(minimizer=LikelihoodOptimizer)

        expected = {
            "c_A": 0.1940912954966998,
            "c_B": 0.5552712770246595,
            "c_C": 0.25063742747864065,
            "mu_B": 0.5418477340418327,
            "mu_C": 0.9177305634612395,
            "sigma_A": 0.08718283966513221,
            "sigma_C": 0.10072652395544755,
        }
        for k in expected:
            assert result.parameters[k] == pytest.approx(expected[k], rel=0.1)

        for fixed_param in ["mu_A", "sigma_B"]:
            assert result.fixed_parameters[fixed_param] == guess[fixed_param]

    def test_global_gmm(self):
        """Test fitting of multiple GMM datasets with overlapping populations"""
        clear_symbols()
        np.random.seed(43)

        states = ["ABC", "BCD"]

        gt = {
            "mu_A": 0.23,
            "mu_B": 0.55,
            "mu_C": 0.92,
            "mu_D": 0.32,
            "sigma_A": 0.1,
            "sigma_B": 0.1,
            "sigma_C": 0.1,
            "sigma_D": 0.2,
            "c_A": 0.22,
            "c_B": 0.53,
            "c_C": 0.25,
            "c_D": 0.22,
        }

        vars = ["x1", "x2"]
        data = {}
        Ns = [1000, 1500]
        for st, var, N in zip(states, vars, Ns):
            data[var] = np.concatenate(
                [
                    np.random.normal(
                        loc=gt[f"mu_{s}"], scale=gt[f"sigma_{s}"], size=int(N * gt[f"c_{s}"]),
                    )
                    for s in st
                ]
            ).reshape(-1, 1)

        # todo guess is not used
        guess = {
            "mu_A": 0.2,
            "mu_B": 0.4,
            "mu_C": 0.7,
            "mu_D": 0.15,
            "sigma_A": 0.1,
            "sigma_B": 0.1,
            "sigma_C": 0.1,
            "sigma_D": 0.1,
            "c_A": 0.33,
            "c_B": 0.33,
            "c_C": 0.33,
            "c_D": 0.33,
        }

        model_dict = {}

        states = ["A", "B", "C"]
        g_shape = (1, 3)
        c_shape = (3, 1)
        mu = symbol_matrix(name="mu", shape=g_shape, suffix=states)
        sigma = symbol_matrix(name="sigma", shape=g_shape, suffix=states)
        c = symbol_matrix(name="c", shape=c_shape, suffix=states)
        model_dict[Symbol("p1")] = Mul(c, GMM(Symbol("x1"), mu, sigma))

        states = ["B", "C", "D"]
        mu = symbol_matrix(name="mu", shape=g_shape, suffix=states)
        sigma = symbol_matrix(name="sigma", shape=g_shape, suffix=states)
        c = symbol_matrix(name="c", shape=c_shape, suffix=states)
        model_dict[Symbol("p2")] = Mul(c, GMM(Symbol("x2"), mu, sigma))

        model = Model(model_dict)

        parameters = Parameters.from_symbols(model.symbols, gt)
        fit = Fit(model, parameters, data, loss=LogSumLoss(sum_axis=1))
        result = fit.execute(minimizer=LikelihoodOptimizer)

        expected = {
            "c_A": 0.2146783006888742,
            "c_B": 0.5418633676210778,
            "c_C": 0.24177790608861852,
            "c_D": 0.21747901002459014,
            "mu_B": 0.5528955016879634,
            "mu_A": 0.2296754306071527,
            "mu_D": 0.3386833210125317,
            "mu_C": 0.9223072648993988,
            "sigma_C": 0.093230280142199,
            "sigma_A": 0.0962655449677717,
            "sigma_B": 0.10288003410897235,
            "sigma_D": 0.21816387196646111,
        }

        for k in expected.keys():
            assert result.parameters[k] == pytest.approx(expected[k], rel=0.1)

    # @pytest.mark.skip("Long execution time")
    def test_markov_gmm(self):
        arr = np.genfromtxt(root_dir / "examples" / "data/GMM_dynamics.txt")
        data = {"e": arr[:, 0].reshape(-1, 1), "t": arr[:, 1]}

        guess_values = {
            "k_A_B": 1e-1,
            "k_B_A": 1e-1,
            "k_B_C": 1e-1,
            "y0_A": 0.6,
            "y0_B": 0.0,
            "mu_A": 0.7,
            "mu_B": 0.05,
            "mu_C": 0.4,
            "sigma_A": 0.1,
            "sigma_B": 0.2,
            "sigma_C": 0.1,
        }

        clear_symbols()
        # np.random.seed(43)

        connectivity = ["A <-> B -> C"]
        m = generate_transition_matrix(connectivity)
        states = extract_states(connectivity)

        # Temporal part
        xt = exp(m * Symbol("t"))

        # Future implementation needs constraints here, sum of y0 should be 1.
        y0 = Matrix([[Symbol("y0_A"), Symbol("y0_B"), 1 - Symbol("y0_A") - Symbol("y0_B")]]).T

        # Gaussian mixture model part
        mu = symbol_matrix("mu", shape=(1, 3), suffix=states)
        sigma = symbol_matrix("sigma", shape=(1, 3), suffix=states)
        gmm = GMM(Symbol("e"), mu=mu, sigma=sigma)

        model = Model({Symbol("p"): Mul(xt @ y0, gmm)})

        parameters = Parameters.from_symbols(model.symbols, guess_values)

        parameters.set("y0_A", lower_bound=0.0, upper_bound=1.0)
        parameters.set("y0_B", lower_bound=0.0, upper_bound=1.0, fixed=True)

        parameters.set("k_A_B", lower_bound=1e-3, upper_bound=1e2)
        parameters.set("k_B_A", lower_bound=1e-3, upper_bound=1e2)
        parameters.set("k_B_C", lower_bound=1e-3, upper_bound=1e2)

        # To calculate the likelihood for a measurement we need to sum the individual probabilities for all states
        # Thus we need to define which axis this is in the model
        STATE_AXIS = 1

        fit = Fit(model, parameters, data, loss=LogSumLoss(sum_axis=STATE_AXIS))
        result = fit.execute(minimizer=LikelihoodOptimizer, max_iter=200, verbose=True,)

        expected = {
            "k_A_B": 0.5415993464686054,
            "k_B_A": 0.08259132883479212,
            "k_B_C": 0.2527748185081457,
            "y0_A": 0.9696231022059791,
            "mu_A": 0.822262354106825,
            "mu_B": 0.12972476836918412,
            "mu_C": 0.5518311456516388,
            "sigma_A": 0.09320689716234987,
            "sigma_B": 0.12251906102401328,
            "sigma_C": 0.07922175330380453,
        }

        print(result.parameters)

        for k in expected:
            assert result.parameters[k] == pytest.approx(expected[k], rel=0.1)



"""
Model operations
Currently only multiplications for probablities
"""
from functools import reduce
from operator import mul

import numpy as np
import numpy.typing as npt

from slimfit.numerical import to_numerical, CompositeExpr
from slimfit.parameter import Parameter
from slimfit.typing import Shape


class CompositeArgsExpr(CompositeExpr):
    """Composite expr which takes *args to init rather than dictionary of expressions"""

    def __init__(self, *args):
        expr = {i: arg for i, arg in enumerate(args)}
        super().__init__(expr)

    def to_numerical(self):
        args = (to_numerical(expr) for expr in self.values())
        instance = self.__class__(*args)

        return instance


class Mul(CompositeArgsExpr):
    # might be subject to renaming
    """Mul elementwise lazily
    """

    def __init__(self, *args):
        super().__init__(*args)

    def __call__(self, **kwargs) -> npt.ArrayLike:
        result = super().__call__(**kwargs)

        return reduce(mul, result.values())

    def __repr__(self):
        args = ", ".join([arg.__repr__() for arg in self.values()])
        return f"Mul({args})"


class MatMul(CompositeArgsExpr):

    """
    matmul composite callable

    arguments must have .shape attribute and shape must be compatible with matrix multiplication

    """

    def __init__(self, *args):
        if len(args) != 2:
            raise ValueError("MatMul takes exactly two arguments")
        super().__init__(*args)

    def __call__(self, **kwargs):
        result = super().__call__(**kwargs)
        return result[0] @ result[1]

    @property
    def shape(self) -> Shape:
        raise NotImplementedError()

    def __repr__(self):
        args = ", ".join([arg.__repr__() for arg in self.values()])
        return f"MatMul({args})"


class Sum(object):
    def __init__(self, matrix, axis=0):
        self.matrix = matrix

    def __call__(self, **kwargs):

        ...



from __future__ import annotations

import re
from collections import UserDict, UserList
from dataclasses import dataclass, field, asdict
from enum import Enum
from functools import cached_property
from typing import Iterable, Optional

import numpy as np
import numpy.typing as npt
from sympy import Expr, Symbol


class ParamType(Enum):
    CONTINUOUS = "continuous"
    DISCRETE = "discrete"
    BOOLEAN = "boolean"


@dataclass(frozen=True)
class Parameter:
    symbol: Expr
    guess: float | int | np.ndarray = field(default=1.0)
    lower_bound: float | int | np.ndarray = field(default=None)
    upper_bound: float | int | np.ndarray = field(default=None)
    # TODO partially fixing an array parameter is not supported
    # perhaps users should use Matrix instead if they want this type of functionality
    fixed: bool = field(default=False)

    def __post_init__(self):

        # If the `guess` has a shape, it must be the same as the symbol shape,
        # if it has any.
        guess_shape = getattr(self.guess, "shape", None)
        symbol_shape = getattr(self.symbol, "shape", guess_shape)
        if guess_shape != symbol_shape:
            raise ValueError(f"Guess shape for symbol {self.symbol} does not match symbol shape")

    @property
    def param_type(self) -> ParamType:
        if "boolean" in self.symbol.assumptions0:
            return ParamType.BOOLEAN
        elif "integer" in self.symbol.assumptions0:
            return ParamType.DISCRETE
        else:
            return ParamType.CONTINUOUS

    @property
    def shape(self) -> tuple[int, ...]:
        """
        Shape of the Parameter. First tries to infer the shape from `Parameter.symbol`, otherwise
        from `Parameter.guess`, and returns an empty tuple if neither is found.

        """
        if shape := getattr(self.symbol, "shape", None):
            return shape
        elif shape := getattr(self.guess, "shape", None):
            return shape

        # when the parameter is a scalar, return an empty tuple, which is the same shape as
        # returned by np.asarray(3.).shape
        return tuple()

    @property
    def name(self) -> str:
        # Do symbols always have names?
        return self.symbol.name


# frozen?
# frozen might not be nessecary but fit should make a copy to prevent modification
class Parameters(UserList):
    """Parameter list object

    or maybe it should be a dict?
    Could potentially help the `Objective` to/from flat array of guesses for argument of scipy.minimize
    """

    @classmethod
    def from_symbols(
        cls,
        symbols: Iterable[Symbol],
        parameters: dict[str, npt.ArrayLike] | Iterable[str] | str = None,
    ) -> Parameters:

        symbol_dict = {symbol.name: symbol for symbol in sorted(symbols, key=str)}

        if isinstance(parameters, str):
            p_list = [Parameter(symbol_dict[k]) for k in re.split("; |, |\*|\s+", parameters)]
        elif isinstance(parameters, list):
            p_list = [Parameter(symbol_dict[k]) for k in parameters]
        elif isinstance(parameters, dict):
            p_list = [Parameter(symbol_dict[k], guess=v) for k, v in parameters.items()]
        elif parameters is None:
            p_list = [Parameter(symbol) for symbol in symbol_dict.values()]
        else:
            raise ValueError("Invalid values for 'parameters' or 'guess'")
        return cls(p_list)

    @property
    def _symbols(self) -> list[Symbol]:
        return [p.symbol for p in self]

    @property
    def _names(self) -> list[str]:
        return [p.name for p in self]

    def index(self, item: Parameter | Symbol | str, *args) -> int:
        # are you really sure it shoulnt be a dict?
        # perhaps not since they 'keys' could be str or symbols ?
        if isinstance(item, Parameter):
            return super().index(item, *args)
        elif isinstance(item, Symbol):
            return self._symbols.index(item, *args)
        else:
            return self._names.index(item, *args)

    def set(self, symbol_or_name: Symbol | str, **kwargs):
        idx = self.index(symbol_or_name)

        # todo sanitize kwargs
        self[idx] = Parameter(**(asdict(self[idx]) | kwargs))

    def update_guess(self, guess: dict[str | Symbol, np.ndarray | float]) -> Parameters:
        """returns a new parameters object where """

        p_out = Parameters(self)
        for identifier, value in guess.items():
            idx = p_out.index(identifier)
            p_out[idx] = Parameter(**(asdict(self[idx]) | dict(guess=value)))

        return p_out

    @property
    def guess(self) -> dict[str, np.ndarray]:
        return {p.name: np.asarray(p.guess) for p in self}

    @property
    def symbols(self) -> set[Symbol]:
        return set(p.symbol for p in self)



from __future__ import annotations

from typing import Optional

import numpy as np
import numpy.typing as npt
from sympy import Symbol, zeros, Expr, MatrixBase
from sympy.core.cache import clear_cache


def get_symbols(*symbolic_objects) -> dict[str, Symbol]:
    """Returns a dictionary of symbols
    if no object is given, only returns FitSymbols
    otherwise returns dict of symbols in the object
    """
    if len(symbolic_objects) == 0:
        return FitSymbol._instances
    else:
        symbols = set()
        for symbolic_object in symbolic_objects:
            if isinstance(symbolic_object, dict):
                symbols = set()
                for entry in itertools.chain(symbolic_object.keys(), symbolic_object.values()):
                    try:
                        # rhs is a sympy `Expr` and has `free_symbols` as a set
                        symbols |= rhs.free_symbols
                    except TypeError:
                        # rhs is a slimfit `NumExpr` and has a `free_symbols` dictionary
                        symbols |= set(rhs.free_symbols.values())
                return
            elif isinstance(symbolic_object, (Expr, MatrixBase)):
                symbols |= symbolic_object.free_symbols
                # return {symbol.name: symbol for symbol in sorted(symbolic_object.free_symbols, key=str)}
            else:
                raise TypeError(f"Invalid type {type(symbolic_object)!r}")

        return {symbol.name: symbol for symbol in sorted(symbols, key=str)}


def symbol_dict(expr: Expr) -> dict[str, Symbol]:
    return {symbol.name: symbol for symbol in sorted(expr.free_symbols, key=str)}


def clear_symbols():
    clear_cache()
    FitSymbol._instances = {}


class FitSymbol(Symbol):
    _instances: dict[str, FitSymbol] = {}

    def __new__(cls, name: str):
        # Bypass the sympy cache
        if name.startswith("__"):
            raise ValueError("Double underscore leading names are limited to internal use.")
        if name in cls._instances:
            obj = cls._instances[name]
        else:
            obj = Symbol.__new__(cls, name)
            cls._instances[name] = obj
        return obj

    def _sympystr(self, printer, *args, **kwargs):
        return printer.doprint(self.name)

    _lambdacode = _sympystr
    _numpycode = _sympystr
    _pythoncode = _sympystr


def symbol_matrix(
    name: Optional[str] = None,
    shape: Optional[tuple[int, ...]] = None,
    names: Optional[npt.ArrayLike] = None,
    suffix: Optional[npt.ArrayLike] = None,
) -> Matrix:

    if shape is None:
        if names is not None:
            shape = (len(names), 1)
        elif suffix is not None:
            shape = (len(suffix), 1)
        else:
            raise ValueError("If 'shape' is not given, must specify 'names' or 'suffix'")

    # Generate names for parameters. Uses 'names' first, then <name>_<suffix> otherwise generates suffices
    # from indices
    if names is None and name is None:
        raise ValueError("Must specify either 'name' or 'names'")
    elif names is None:
        names = np.full(shape, fill_value="", dtype=object)
        if suffix is None:
            for i, j in np.ndindex(shape):
                names[i, j] = f"{name}_{i}_{j}"
        else:
            suffix = np.array(suffix).reshape(shape)
            for i, j in np.ndindex(shape):
                names[i, j] = f"{name}_{suffix[i, j]}"
    else:
        names = np.array(names)

    matrix = zeros(*shape)
    for i, j in np.ndindex(shape):
        matrix[i, j] = Symbol(name=names[i, j],)

    return matrix


# SORT_PRIORITY = [Variable, Probability, Parameter]
# SORT_KEY = lambda x: (SORT_PRIORITY.index(type(x)), x.name)



from __future__ import annotations

from functools import cached_property
from typing import Optional, Type

import numpy as np
import numpy.typing as npt
from sympy import Expr

from slimfit.loss import L2Loss, LogLoss, Loss
from slimfit.minimizer import ScipyMinimizer, Minimizer
from slimfit.models import Model
from slimfit.numerical import to_numerical
from slimfit.parameter import Parameter, Parameters


class Fit(object):
    """fit objects take model and data

    their function is to determine minimizer / loss strategies and execute those

    """

    def __init__(
        self,
        model: Model,
        parameters: list[Parameter] | Parameters,
        data: dict[str | Expr, npt.ArrayLike],
        loss: Optional[Loss] = L2Loss(),
        # posterior: Optional[CompositeNumExpr],
    ):

        self.symbolic_model = model

        # make a new instance such that external modification does not affect the
        # copy stored here
        self.parameters = Parameters(parameters)

        # Convert data keys to `str` if given as `Symbol; data values as arrays
        data: dict[str, np.ndarray] = {
            getattr(k, "name", k): np.asarray(v) for k, v in data.items()
        }
        self.loss = loss

        # 'independent' data; or 'xdata'; typically chosen measurement points
        self.xdata = {k: v for k, v in data.items() if k in self.symbolic_model.symbol_names}

        # 'dependent' data; or 'ydata'; typically measurements
        self.ydata = {k: v for k, v in data.items() if k in self.symbolic_model.dependent_symbols}

    def execute(
        self, minimizer: Optional[Type[Minimizer]] = None, **execute_options,
    ):

        minimizer_cls = minimizer or self.get_minimizer()
        minimizer_instance = minimizer_cls(
            self.symbolic_model.to_numerical(), self.parameters, self.loss, self.xdata, self.ydata
        )

        result = minimizer_instance.execute(**execute_options)

        result.symbolic_model = self.symbolic_model

        return result

    def get_minimizer(self) -> Type[Minimizer]:
        """Automatically determine which minimizer to use"""
        return ScipyMinimizer

    def get_loss(self, **kwargs):
        raise NotImplementedError()
        if self.model.probabilistic:
            return LogLoss(**kwargs)
        else:
            return L2Loss(**kwargs)



from __future__ import annotations

import itertools
from typing import Union, ItemsView, ValuesView, KeysView, Optional

import numpy.typing as npt
from sympy import Expr, MatrixBase, Symbol

import slimfit.numerical as numerical
from slimfit.parameter import Parameters


class Model(numerical.CompositeExpr):
    def __init__(
        self, model_dict: dict[Symbol | str, Expr | numerical.NumExprBase | MatrixBase],
    ):

        # currently typing has a small problem where keys are expected to be `str`, not symbol
        super().__init__(model_dict)

    def __repr__(self):
        return f"Model({self.expr.__repr__()})"

    # def __getitem__(self, item: Union[str, Symbol]) -> numerical.NumExprBase:
    #     if isinstance(item, str):
    #         item = self.symbols[item]
    #
    #     return self.expr[item]

    @property
    def dependent_symbols(self) -> dict[str, Symbol]:
        # todo needs to be updated
        """Variables corresponding to dependent (measured) data, given as keys in the model dict"""

        return {symbol.name: symbol for symbol in self.expr.keys()}



import setuptools
import versioneer
import sys
import os

# ensure the current directory is on sys.path so versioneer can be imported
# when pip uses PEP 517/518 build rules.
# https://github.com/python-versioneer/python-versioneer/issues/193
sys.path.append(os.path.dirname(__file__))

if __name__ == "__main__":
    setuptools.setup(
        version=versioneer.get_version(),
        cmdclass=versioneer.get_cmdclass(),
    )



from __future__ import annotations


from typing import Iterable

import numpy as np

from slimfit import Model
from slimfit.loss import Loss
from slimfit.typing import Shape

MIN_PROB = 1e-9  # Minimal probability value (> 0.) to enter into np.log


# py3.10:
# from dataclasses import dataclass, field
# from functools import cached_property
# slots=True,
# kw_only = True
# @dataclass(frozen=True)


class Objective:
    def __init__(
        self,
        model: Model,
        loss: Loss,
        xdata: dict[str, np.ndarray],
        ydata: dict[str, np.ndarray],
        negate: bool = False,
    ):
        self.model = model
        self.loss = loss
        self.xdata = xdata
        self.ydata = ydata

        self.sign = -1 if negate else 1


class ScipyObjective(Objective):
    def __init__(
        self,
        model: Model,
        loss: Loss,
        xdata: dict[str, np.ndarray],
        ydata: dict[str, np.ndarray],
        shapes: dict[str, Shape],
        negate: bool = False,
    ):
        super().__init__(model=model, loss=loss, xdata=xdata, ydata=ydata, negate=negate)
        self.shapes = shapes

    def __call__(self, x: np.ndarray) -> float:
        parameters = unpack(x, self.shapes)

        y_model = self.model(**parameters, **self.xdata)
        loss = self.loss(self.ydata, y_model)

        return self.sign * loss


class ScipyEMObjective(Objective):
    def __init__(
        self,
        model: Model,
        loss: Loss,
        xdata: dict[str, np.ndarray],
        posterior: dict[str, np.ndarray],
        shapes: dict[str, Shape],
        negate: bool = False,  # todo actually use the negate bool
    ):
        super().__init__(model=model, loss=loss, xdata=xdata, ydata={}, negate=negate)
        self.posterior = posterior
        self.shapes = shapes

    def __call__(self, x: np.ndarray) -> float:
        parameters = unpack(x, self.shapes)

        probability = self.model(**parameters, **self.xdata)

        # Todo do this in a `loss`
        expectation = {
            lhs: self.posterior[lhs] * np.log(np.clip(prob, a_min=MIN_PROB, a_max=1.0))
            for lhs, prob in probability.items()
        }

        # TODO: LOSS / WEIGHTS

        return -sum(r.sum() for r in expectation.values())


# seperate functions?
def unpack(x: np.ndarray, shapes: dict[str, Shape]) -> dict[str, np.ndarray]:
    """Unpack a ndim 1 array of concatenated parameter values into a dictionary of
        parameter name: parameter_value where parameter values are cast back to their
        specified shapes.
    """
    sizes = [int(np.product(shape)) for shape in shapes.values()]

    x_split = np.split(x, np.cumsum(sizes))
    p_values = {name: arr.reshape(shape) for (name, shape), arr in zip(shapes.items(), x_split)}

    return p_values


def pack(parameter_values: Iterable[np.ndarray]) -> np.ndarray:
    """Pack a dictionary of parameter_name together as array"""

    return np.concatenate(tuple(param_value.ravel() for param_value in parameter_values))



from __future__ import annotations

import abc
from typing import Optional, Literal

import numpy as np
import numpy.typing as npt

from slimfit.reduce import (
    ReductionStrategy,
    mean_reduction,
    sum_reduction,
    concat_reduction,
)


class Loss(object):
    def __init__(
        self,
        weights: Optional[dict[str, npt.ArrayLike]] = None,
        reduction: Literal["mean", "sum", "concat", "none", None] = "mean",
    ):

        self.weights = weights
        if reduction == "mean":
            self.reduce: ReductionStrategy = mean_reduction
        elif reduction == "sum":
            self.reduce: ReductionStrategy = sum_reduction
        elif reduction == "concat":
            self.reduce: ReductionStrategy = concat_reduction
        elif reduction in [None, "none"]:
            self.reduce = lambda x: x

    @abc.abstractmethod
    def __call__(
        self, y_data: dict[str, np.ndarray], y_model: dict[str, np.ndarray]
    ) -> np.ndarray | float:
        ...

    # def reduce(
    #     self, residuals: dict[str, np.ndarray]
    # ) -> float | np.ndarray | dict[str, np.ndarray]:
    #     if self.reduction == "mean":
    #         size = sum(arr.size for arr in residuals.values())
    #         total = sum(r.sum() for r in residuals.values())
    #         return total / size
    #     elif self.reduction == "sum":
    #         return sum(r.sum() for r in residuals.values())
    #     elif self.reduction == "concat":
    #         return np.concatenate([a.ravel() for a in residuals.values()])
    #     elif self.reduction in ["none", None]:
    #         return residuals


# https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.least_squares.html#scipy.optimize.least_squares
# https://pytorch.org/docs/stable/generated/torch.nn.MSELoss.html
class L1Loss(Loss):
    """L1 loss """

    def __call__(
        self, dependent_data: dict[str, np.ndarray], target_data: dict[str, np.ndarray]
    ) -> np.ndarray | float:

        if self.weights is None:
            residuals = {k: (target_data[k] - dependent_data[k]) for k in target_data.keys()}
        else:
            residuals = {
                k: (target_data[k] - dependent_data[k]) * self.weights[k]
                for k in target_data.keys()
            }

        return self.reduce(residuals)


class L2Loss(Loss):
    def __call__(
        self, y_data: dict[str, np.ndarray], y_model: dict[str, np.ndarray]
    ) -> np.ndarray | float:

        if self.weights is None:
            residuals = {k: (y_model[k] - y_data[k]) ** 2 for k in y_model.keys()}
        else:
            residuals = {
                k: ((y_model[k] - y_data[k]) * self.weights[k]) ** 2 for k in y_model.keys()
            }

        return self.reduce(residuals)


# TODO base class for probability-based losses which only take target data in their __call__ (= MLELoss ? )
# then the base should take weights and only do reduction; subclasses to logs / sums etc
MIN_PROB = 1e-9  # Minimal probability value (> 0.) to enter into np.log


class LogLoss(Loss):
    """Takes the elementwise logarithm of predicted input data

    Used in combination with maximum likelihood methods

    returns negative of the reductions are use in combination with minimizers rather than maximizers
    """

    def __init__(
        self,
        weights: Optional[dict[str, npt.ArrayLike]] = None,
        reduction: Literal["mean", "sum", "concat"] = "sum",
    ):
        if reduction not in ["mean", "sum", "concat"]:
            raise ValueError(
                f"LogLoss does not support reduction {reduction!r}, only 'mean', 'sum', 'concat'"
            )
        super().__init__(weights, reduction)

    def __call__(
        self, y_data: dict[str, np.ndarray], y_model: dict[str, np.ndarray]
    ) -> np.ndarray | float:
        if self.weights is None:
            # log_vals = {k: np.log(y_model[k]) for k in y_model.keys()}
            log_vals = {k: np.log(np.clip(y_model[k], a_min=MIN_PROB, a_max=None)) for k in y_model.keys()}

        else:
            log_vals = {k: np.log(y_model[k] * self.weights[k]) for k in y_model.keys()}

        return -self.reduce(log_vals)


class LogSumLoss(Loss):
    """Sums by specified axis, then takes elementwise log, then applies reduction method



    Used in combination with maximum likelihood methods

    Example:
        # sum along axis 1, then takes elementwise log, then sums the result
        LogSumLoss(sum_axis=1, reduction='sum')

    returns negative of the reductions are use in combination with minimizers rather than maximizers
    """

    def __init__(
        self,
        weights: Optional[dict[str, npt.ArrayLike]] = None,
        sum_axis: Optional[int] = 1,
        reduction: Literal["mean", "sum", "concat"] = "sum",
    ):
        self.sum_axis = sum_axis
        if reduction not in ["mean", "sum", "concat"]:
            raise ValueError(
                f"LogSumLoss does not support reduction {reduction!r}, only 'mean', 'sum', 'concat'"
            )
        super().__init__(weights, reduction)

    def __call__(
        self, y_data: dict[str, np.ndarray], y_model: dict[str, np.ndarray]
    ) -> np.ndarray | float:

        # from slimfit.minimizer import MIN_PROB
        if self.weights is None:
            log_vals = {
                k: np.log(
                    #y_model[k].sum(axis=self.sum_axis),
                    np.clip(y_model[k].sum(axis=self.sum_axis), a_min=MIN_PROB, a_max=None)
                )
                for k in y_model.keys()
            }

        else:
            log_vals = {
                k: np.log(
                    # np.clip(y_model[k].sum(axis=self.sum_axis)) * self.weights[k], a_min=MIN_PROB, a_max=None)
                    y_model[k].sum(axis=self.sum_axis)
                    * self.weights[k]
                )
                for k in y_model.keys()
            }

        return -self.reduce(log_vals)


#
#
# class LogLoss(Loss):
#     """Takes the elementwise logarithm of predicted input data
#
#     Used in combination with maximum likelihood methods
#
#     returns negative of the reductions are use in combination with minimizers rather than maximizers
#     """
#
#     def __init__(
#         self,
#             sum_axis: int,
#             weights: Optional[dict[str, npt.ArrayLike]] = None,
#     ):
#         if reduction not in ["mean", "sum", "concat"]:
#             raise ValueError(
#                 f"LogLoss does not support reduction {reduction!r}, only 'mean', 'sum', 'concat'"
#             )
#         super().__init__(weights, reduction)
#
#     def __call__(
#         self, ydata: dict[str, np.ndarray], y_model: dict[str, np.ndarray]
#     ) -> np.ndarray | float:
#         if self.weights is None:
#             log_vals = {k: np.log(y_model[k]) for k in y_model.keys()}
#
#         else:
#             log_vals = {
#                 k: np.log(y_model[k] * self.weights[k]) for k in y_model.keys()
#             }
#
#         return -self.reduce(log_vals)



from __future__ import annotations

from collections import defaultdict
from typing import Iterable, Optional, OrderedDict, Any

import numpy as np
from sympy import Symbol

from slimfit import NumExprBase
from slimfit.models import Model
from slimfit.operations import Mul

# from slimfit.operations import Mul
from slimfit.parameter import Parameter


def overlapping_model_parameters(
    model_callables: list[tuple[Symbol, NumExprBase]], symbols: set[Symbol],
) -> list[Model]:

    seen_models = []
    seen_sets = []
    for lhs, num_expr in model_callables:
        param_set = num_expr.symbols & symbols
        # param_set = set(num_expr.free_parameters.keys())

        found = False
        # look for sets of parameters we've seen so far, if found, append to the list of sets
        for i, test_set in enumerate(seen_sets):
            if param_set & test_set:
                # add additional items to this set of parameters
                test_set |= param_set
                seen_models[i].append((lhs, num_expr))
                found = True
        if not found:
            seen_sets.append(param_set)
            seen_models.append([(lhs, num_expr)])

    # Next, piece together the dependent model parts as Model objects, restoring original multiplications
    sub_models = []
    for components in seen_models:
        model_dict = defaultdict(list)
        for lhs, rhs in components:
            model_dict[lhs].append(rhs)

        model_dict = {
            lhs: rhs[0] if len(rhs) == 1 else Mul(*rhs) for lhs, rhs in model_dict.items()
        }
        sub_models.append(Model(model_dict))

    return sub_models


def get_bounds(
    parameters: Iterable[Parameter],
) -> Optional[list[tuple[Optional[float], Optional[float]]]]:
    """
    Get bounds for minimization.
    Args:
        parameters: Iterable of Parameter objects.

    Returns:
        Either a list of tuples to pass to `scipy.minimize` or None, if there are no bounds.
    """
    bounds = [(p.vmin, p.vmax) for p in parameters]

    if all([(None, None) == b for b in bounds]):
        return None
    else:
        return bounds


def clean_types(d: Any) -> Any:
    """cleans up nested dict/list/tuple/other `d` for exporting as yaml

    Converts library specific types to python native types, including numpy dtypes,
    OrderedDict, numpy arrays

    # https://stackoverflow.com/questions/59605943/python-convert-types-in-deeply-nested-dictionary-or-array

    """
    if isinstance(d, np.floating):
        return float(d)

    if isinstance(d, np.integer):
        return int(d)

    if isinstance(d, np.ndarray):
        return d.tolist()

    if isinstance(d, list):
        return [clean_types(item) for item in d]

    if isinstance(d, tuple):
        return tuple(clean_types(item) for item in d)

    if isinstance(d, OrderedDict):
        return clean_types(dict(d))

    if isinstance(d, dict):
        return {k: clean_types(v) for k, v in d.items()}

    else:
        return d
